
#include"hash.h"

void *hsearch(void *base,int width,int size,unsigned key,match_fn match) {
          int pos,initial,found;
          void *element,*null;

          null = NULL;
          initial = pos = hash(key,size);
          do {
             if(! (found = match(element = (char *)base + (pos*width))) ) {
                if(null != NULL) {
                   memswap(element,null,width);
                   return null;
                }
                return element;
             }
             if(found < 0 && null == NULL) null = element;
          }
          while( ((1 << (++hashes)) <= size) && ((pos = rehash(key,pos,size)) != initial) );
          return NULL;
      }

void *kihsearch(void *base,int width,int size,unsigned key,match_fn match) {
         void *element, *pelement;
         int prev,pprev,curr;

         pelement = NULL;
         hashes = 0;
         prev = -1;
         curr = hash(key,size);
         do {
            ++hashes;
            if( match(element = (char *)base + (curr * width)) ) {
               if(pelement) {
                   memswap(pelement,element,width);
                   return pelement;
               }
               else return element;
            }
            pelement = element;
            pprev = prev;
            prev = curr;
            curr = kirehash(key,curr,hashes,size);
        } while( curr != prev && curr != pprev );
        return NULL;
   }

void **hresize(void **base,int width,int *size,int newsize,hsearch_fn search,key_fn key,match_fn match,fail_fn fail) {
          int oldsize = *size;
          void *element = *base, *newbase = hnew(newsize,width);

          if(newbase == NULL) return NULL;
          while(oldsize > 0) {
               void *newloc = search(newbase,width,newsize,key(element),match);
               if(newloc == NULL) {
                   if(fail(element)) {
                      errno = HRESIZE;
                      free(newbase);
                      return NULL;
                   }
               }
               else memcpy(newloc,element,width);
               oldsize--;
               element = (char *)element + width;
           }
           hdispose(*base);
           *base = newbase;
           *size = newsize;
           return base;
    }

unsigned str2key(char *string) {
           unsigned key = 1;
           char prev = *string;

           while(*string) {
              key *= (*string + prev);
              key /= (abs(prev - *string) + 1);
              key += *string;
              prev = *string;
              string++;
           }
           return key;
        }

void memswap(void *loc_0,void *loc_1,int size) {
         while(size > sizeof(unsigned))  {
             unsigned temp = *(unsigned *)loc_0;
             *(unsigned *)loc_0 = *(unsigned *)loc_1;
             *(unsigned *)loc_1 = temp;
             (unsigned *)loc_0++;
             (unsigned *)loc_1++;
             size -= sizeof(unsigned);
         }
         while(size-- > 0) {
            char temp = *(char *)loc_0;
             *(char *)loc_0 = *(char *)loc_1;
             *(char *)loc_1 = temp;
             (char *)loc_0++;
             (char *)loc_1++;
         }
         return;
    }

#define ctrans(c) (islower(c) ? (c) - 'a' : isupper(c) ? 26 + (c) - 'A' : isdigit(c) ? 52 + (c) - '0' : (-1))
#define citrans(c) (isalpha(c) ? tolower(c) - 'a' : isdigit(c) ? 26 + (c) - '0' : (-1))

#define ALPHA_DIVISOR  (62 << 2)
#define IALPHA_DIVISOR (36 << 2)
#define BYTE_DIVISOR 0x3f

int ostrhash(char *string,int size) {
         unsigned addr = 0,alpha = OSTRHASH_LEN;

         while(*string && alpha > 0) {
            char letter = ctrans(*string);
            if(letter >= 0) {
                addr |= ((unsigned)(letter / ALPHA_DIVISOR)) << ((alpha << 2) - 2);
                alpha--;
            }
            string++;
         }
         return ohash(addr,size);
    }

int ostrihash(char *string,int size) {
             unsigned addr = 0,alpha = OSTRHASH_LEN;

             while(*string && alpha > 0) {
                char letter = citrans(*string);
                if(letter >= 0) {
                    addr |= ((unsigned)(letter / IALPHA_DIVISOR)) << ((alpha << 2) - 2);
                    alpha--;
                }
                string++;
             }
             return ohash(addr,size);
         }

int   obstrhash(unsigned char *string,int size) {
             unsigned addr = 0,byte = 0;

             while(byte < OSTRHASH_LEN) {
                addr <<= 2;
                addr |=  (string[byte] / BYTE_DIVISOR);
                byte++;
             }
             return ohash(addr,size);
      }
